
/*
  WiFi connected Woord LED Clock. It gets NTP time from the internet and translates to two CJMCU 64 Bit WS2812 5050 RGB LED Driver Development Board

  see comment tab for configuration and explanation

*/

// front cover version
#define versie 3

// memory
#include <EEPROM.h>   // memory

// Wifi
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>

// Fastled
#include <FastLED.h>
// #define DEBUG_ON

FASTLED_USING_NAMESPACE

#if defined(FASTLED_VERSION) && (FASTLED_VERSION < 3001000)
#warning "Requires FastLED 3.1 or later; check github for latest code."
#endif

// wifi credentials
#include "Secrets.h"             // this file is intentionally not included in the sketch, so nobody accidentally commits their secret information.
// create a Secrets.h file in the same directory as your schets with the following:
// #define WIFI_SSID "your-ssid"
// #define WIFI_PASS "your-password"
const char ssid[] = WIFI_SSID;        // Your network SSID name here
const char pass[] = WIFI_PASS;    // Your network password here

// time
unsigned long timeZone = 1.0;    // Change this value to your local timezone (in my case +1 for Amsterdam)

ESP8266WiFiMulti wifiMulti;
WiFiUDP UDP;
IPAddress timeServerIP;
const char* NTPServerName = "time.nist.gov";
const int NTP_PACKET_SIZE = 48;
byte NTPBuffer[NTP_PACKET_SIZE];

unsigned long intervalNTP = 5 * 60000; // Request NTP time every 5 minutes
unsigned long prevNTP = 0;
unsigned long lastNTPResponse = millis();
// uint32_t timeUNIX = 1638701389;   // 10 voor twaalf just to make a start time in case no NTP responce. was 0
uint32_t timeUNIX = 1638701989;
unsigned long prevActualTime = 0;

#define LEAP_YEAR(Y) ( ((1970+Y)>0) && !((1970+Y)%4) && ( ((1970+Y)%100) || !((1970+Y)%400) ) )
static const uint8_t monthDays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

// fastled
#define VOLTS          5
#define MAX_MA       500
uint8_t gHue = 32; // rotating "base color" used by many of the patterns
#define MAX_BRIGHTNESS  200
#define MIN_BRIGHTNESS  20
#define BRIGHTNESS_STEP 15

#define FRAMES_PER_SECOND   100   // was 120

#define LED_TYPE    WS2812
#define COLOR_ORDER GRB
#define NUM_LEDS 128
#define DATA_PIN 2 // D4 
CRGB LEDs[NUM_LEDS];
byte Map[NUM_LEDS] ;  // map matrix arangement

byte mapindex[NUM_LEDS]; // mapped leds for color effects to apply
byte maplength;  // length of sentence

// button and menu
#define BUTTON_PIN 0 // D3 pull-up
bool Buttonpressed = false;
unsigned long Buttontimer = 0;
unsigned long Nopresstimer = 0;
#define LONGPRESS 700   // millis time to enter menu
#define WAITPRESS 1500  // millis time to release button
#define NOPRESS 5000   // millis time to leave menu
byte menulevel = 0;
byte menu = 0;
byte menukeuze = 0;
bool firstentry = true;
byte randomcoloreffect = 0;

// memory (place) place increment depending on data type number of bytes
byte brightness = 100;      // (0) 0 - 255
byte setvolume = 20;        // (1) 10 - 30  volume (step 5)
byte setEQ = 0;             // (2) 0 - 5  klankkleur
uint8_t deltaHue = 0;       // (3) 0 - 35 delta hue between rainbow colored leds 0= no rainbow (max 35, step 7)
byte timesignal = 2;        // (4) 0 = no and 1= oud 2 = koekoek 3 = paard 4 = pling
byte timecoloreffect = 1;   // (5) 0 = no 1= random, 2= glitter, 3= hartslag, 4=sinusgolf, 5=jongleren, 6=confetti

struct DateTime {
  int  year;
  byte month;
  byte day;
  byte hour;
  byte minute;
  byte second;
  byte dayofweek;
};

DateTime currentDateTime;

/* DFPlayer */
#include "Arduino.h"
#include "SoftwareSerial.h"
#include "DFPlayerMini_Fast.h"  // https://github.com/PowerBroker2/DFPlayerMini_Fast   note; .isPlaying() function does not work! so use busy pin!

// SoftwareSerial ssDFPlayer = SoftwareSerial(12, 14); // RX to TX, TX to RX Arduino uno
SoftwareSerial ssDFPlayer(12, 14); // RX to TX, TX to RX D1 mini
DFPlayerMini_Fast DFPlayer;
#define DFPlayerBusyPin 13 // Low means playing, High means not playing
int stepvolume = 5;  // default step incr of decr volume


void setup() {

  Serial.begin(115200);
  delay(3000);
  Serial.println("setup");

  EEPROM.begin(512);
  // read saved values, if there aren't write them to memory
  Serial.println("reading EEPROM");
  if (EEPROM.read(0) == 0) {           // indien memory empty
    EEPROM.put(0, brightness) ;                 // write default to memory
  }
  else EEPROM.get(0, brightness);              // write memory to default

  if (EEPROM.read(1) == 0) {
    EEPROM.put(1, setvolume);
  }
  else EEPROM.get(1, setvolume);

  if (EEPROM.read(2) == 0) {
    EEPROM.put(2, setEQ);
  }
  else EEPROM.get(2, setEQ);

  if (EEPROM.read(3) == 0) {
    EEPROM.put(3, deltaHue);
  }
  else EEPROM.get(3, deltaHue);

  if (EEPROM.read(4) == 0) {
    EEPROM.put(4, timesignal);
  }
  else EEPROM.get(4, timesignal);

  if (EEPROM.read(5) == 0) {
    EEPROM.put(5, timecoloreffect);
  }
  else EEPROM.get(5, timecoloreffect);

  Serial.print("brightness: "); Serial.println(brightness);
  Serial.print("volume: "); Serial.println(setvolume);
  Serial.print("klankkleur: "); Serial.println(setEQ);
  Serial.print("delta Hue: "); Serial.println(deltaHue);
  Serial.print("tijd signaal: "); Serial.println(timesignal);
  Serial.print("kleur effect: "); Serial.println(timecoloreffect);

  EEPROM.commit();

  // start DFPlayer
  ssDFPlayer.begin(9600);

  Serial.println(F("Initializing DFPlayer ... (May take 3~5 seconds)"));
  DFPlayer.begin(ssDFPlayer, false);  //bool begin(Stream& stream, bool debug, unsigned long threshold=100);

  Serial.println(F("DFPlayer Mini online."));
  DFPlayer.setTimeout(500); //Set serial communictaion time out 500ms
  DFPlayer.volume(setvolume);  //Set volume value
  delay(100);
  DFPlayer.EQSelect(setEQ);
  delay(100);

  pinMode(DFPlayerBusyPin, INPUT);

  speeltrack(1, true);
  speeltrack(2, true);

  FastLED.delay(3000);
  FastLED.setMaxPowerInVoltsAndMilliamps( VOLTS, MAX_MA);
  // FastLED.addleds
  FastLED.addLeds<LED_TYPE, DATA_PIN, COLOR_ORDER>(LEDs, NUM_LEDS).setCorrection(TypicalSMD5050);
  FastLED.setBrightness(brightness);
  // FastLED.addLeds<WS2812B, DATA_PIN, GRB>(LEDs, NUM_LEDS);

  //configure pin 0 as an input and enable the internal pull-up resistor
  pinMode(BUTTON_PIN, INPUT_PULLUP);

  startWiFi();
  startUDP();

  if (!WiFi.hostByName(NTPServerName, timeServerIP)) {
    Serial.println("DNS lookup failed. Rebooting.");
    speeltrack(3, true);
    Serial.flush();
    ESP.reset();
  }
  Serial.print("Time server IP:\t");
  Serial.println(timeServerIP);

  Serial.println("Sending NTP request ...");
  sendNTPpacket(timeServerIP);

  // fill Map[] Map contains array index sequence to matrix orientation
  Serial.println("Fill Map");
  for (int r = 0 ; r < 8 ; r++) {
    for (int m = 0 ; m < 2; m++) {
      for (int k = 0; k < 8; k++) {
        Map[r * 16 + m * 8 + k] = r * 8 + m * 64 + k;
        //        Serial.print("index: ");
        //        Serial.print(r * 16 + m * 8 + k);
        //        Serial.print(" led: ");
        //        Serial.println(r * 8 + m * 64 + k);
      }
    }
  }

}

void loop() {
  unsigned long currentMillis = millis();

  // acquire time
  if (currentMillis - prevNTP > intervalNTP) { // If a interval has passed since last NTP request
    prevNTP = currentMillis;
    Serial.println("\r\nSending NTP request ...");
    sendNTPpacket(timeServerIP);               // Send an NTP request
  }

  uint32_t time = getTime();                   // Check if an NTP response has arrived and get the (UNIX) time
  if (time) {                                  // If a new timestamp has been received
    timeUNIX = time;
    Serial.print("NTP response:\t");
    Serial.println(timeUNIX);
    lastNTPResponse = currentMillis;
  } else if ((currentMillis - lastNTPResponse) > 3600000) {
    Serial.println("More than 1 hour since last NTP response. Rebooting.");
    speeltrack(4, true);
    Serial.flush();
    ESP.reset();
  }
  // menu handling
  int Button = digitalRead(BUTTON_PIN);
  if (Button == LOW && !Buttonpressed) {    // it means it is pressed
    Buttonpressed = true;
    Buttontimer = currentMillis;   //start timer
  }
  if (Button == LOW && Buttonpressed) {
    if (menulevel <= 1 && currentMillis - Buttontimer >= LONGPRESS) {
      // do longpress action
      // enter menu and toggle menu
      speeltrack(6, true);  // menukeuze
      menulevel = 1;
      menu++;
      if (menu > 6) menu = 1;
      if (menu < 6) speeltrack(6 + menu, true);   // 1=helderheid, 2=regenboog, 3=volume, 4=tijdsignaal, 5=tijdeffect
      else speeltrack(104, true);   // 6=klankkleur
      delay(WAITPRESS);
    } else if (menulevel == 2) {  // toggle submenu
      switch (menu) {
        case 1:  // helderheid
          if (firstentry) {
            menukeuze = brightness;
          } else {
            menukeuze += BRIGHTNESS_STEP;
            if (menukeuze > MAX_BRIGHTNESS) menukeuze = MIN_BRIGHTNESS;
            FastLED.setBrightness(menukeuze);
            delay(10);
            //          FastLED.show();
            //          FastLED.delay(100);
          }
          speeltrack(7, true);
          speeltrack(1000 + ((menukeuze - MIN_BRIGHTNESS) / BRIGHTNESS_STEP), true);
          break;
        case 2:   // regenboog
          if (firstentry) {
            menukeuze = deltaHue;
          } else {
            menukeuze += 7;    // increase delta hue step
            if (menukeuze > 35) menukeuze = 0;
            rainbow();
            //            FastLED.show();
            //            FastLED.delay(100);
          }
          speeltrack(8, true);
          speeltrack(1000 + (menukeuze / 7), true);
          break;
        case 3:   // volume
          if (firstentry) {
            menukeuze = setvolume;
          } else {
            menukeuze += stepvolume;  // increase vol by
            if (menukeuze > 30) menukeuze = 10;
            DFPlayer.volume(menukeuze);  //Set volume value
            delay(100);
          }
          speeltrack(9, true);
          speeltrack(1000 + (menukeuze / stepvolume), true);
          break;
        case 4:    // tijdsignaal
          if (firstentry) {
            menukeuze = timesignal;
          } else {
            menukeuze += 1;  // increase
            if (menukeuze > 4) menukeuze = 0;
          }
          speeltrack(10, true);
          if (menukeuze == 0) {
            speeltrack(21, true);   // uit
          } else {
            speeltrack(315 + menukeuze * 5, true);
          }
          break;
        case 5:  // tijdkleureffect
          if (firstentry) {
            menukeuze = timecoloreffect;
          } else {
            menukeuze += 1;  // increase
            if (menukeuze > 6) menukeuze = 0;
          }
          speeltrack(11, true);
          if (menukeuze == 0) {
            speeltrack(21, true);   // uit
          } else {
            speeltrack(11 + menukeuze, true);   // kwartier
          }
          break;
        case 6: // klankkleur
          if (firstentry) {
            menukeuze = setEQ;
          } else {
            menukeuze += 1; // increase
            if (menukeuze > 5) menukeuze = 0;
            DFPlayer.EQSelect(menukeuze);
            delay(100);
          }
          speeltrack(105 + menukeuze, true);
          break;
        default:
          // do nothing
          break;
      }
      firstentry = false;
      delay(WAITPRESS);
    }
    // else do nothing button still pressed
  }

  if (Button == HIGH && Buttonpressed) {    // it is released
    Buttonpressed = false;
    if (menulevel == 0 && currentMillis - Buttontimer < LONGPRESS) { // it was a short press
      // speak time
      speaktime();
    } else if (menulevel == 1) {   // it was a long press and menulevel 1 choice is made goto level 2
      menulevel = 2;   // raise to submenu
      Nopresstimer = currentMillis;   // set no press timer
      firstentry = true;
    } else if (menulevel == 2) {  // a choice was made
      // write choice to memory
      switch (menu) {
        case 1:     // helderheid
          brightness = menukeuze;
          EEPROM.put(0, brightness) ;
          Serial.print("brightness: "); Serial.print(brightness);
          break;
        case 2: // regenboog
          deltaHue = menukeuze;
          EEPROM.put(3, deltaHue) ;
          Serial.print("Delta Hue: "); Serial.print(deltaHue);
          break;
        case 3: // volume
          setvolume = menukeuze;
          EEPROM.put(1, setvolume) ;
          Serial.print("Volume: "); Serial.print(setvolume);
          break;
        case 4:  // tijdsignaal
          timesignal = menukeuze;
          EEPROM.put(4, timesignal) ;
          Serial.print("tijd signaal: "); Serial.print(timesignal);
          break;
        case 5: // tijd kleureffect
          timecoloreffect = menukeuze;
          EEPROM.put(5, timecoloreffect) ;
          Serial.print("kleur effect: "); Serial.print(timecoloreffect);
          break;
        case 6: // klankkleur
          setEQ = menukeuze;
          EEPROM.put(2, setEQ) ;
          Serial.print("klank kleur: "); Serial.print(setEQ);
          break;
        default:
          break;
      }  // leave menu
      EEPROM.commit();
      Serial.println(" wrote to memory");
      //      menukeuze = 0;
      //     menu = 0;
      menulevel = 0;
      Nopresstimer = 0;
    }
  }

  if (Button == HIGH && Nopresstimer != 0 && currentMillis - Nopresstimer >= NOPRESS) {   // exit menu if no choice is made
    //    menukeuze = 0;
    //    menu = 0;
    speeltrack(5, true);  // instructie
    menulevel = 0;
    Nopresstimer = 0;
  }

  // time handling
  uint32_t actualTime = timeUNIX + (currentMillis - lastNTPResponse) / 1000;
  if (actualTime != prevActualTime && timeUNIX != 0) { // If a second has passed since last update

    prevActualTime = actualTime;
    convertTime(actualTime);

    for (int i = 0; i < NUM_LEDS; i++)
      LEDs[i] = CRGB::Black;

    prepareLEDtime();
    applyMap();
    rainbow();
  }

  // apply time sound and/or color effect
  byte hours = currentDateTime.hour;
  byte seconds = currentDateTime.second;
  byte minutes = currentDateTime.minute;
  // apply speach first
  if (minutes == 0 && seconds == 0 && timesignal > 0 && hours>=6 && hours<=22) {
    //  if (seconds == 0) {
    // hour effect
    speeltrack(320 + timesignal, true);
    delay(1000); // prevent multiple sound effect within a second
//    if (hours > 12) hours = hours - 12;
//    for (int i = 0; i < hours; i++) {
//      speeltrack(323, true);
//    }
    
  }

  // apply time color effects for a minute

  if (timecoloreffect > 0) {
    if (minutes == 0 || minutes == 15 || minutes == 30 || minutes == 45)  {
      if (timecoloreffect == 1 && randomcoloreffect == 0) {
        randomcoloreffect = random8(2, 7);
      } else {
        randomcoloreffect = timecoloreffect;
      }
      applycoloreffect(timecoloreffect);
    } else {
      randomcoloreffect = 0;
    }
  }


  FastLED.show();
  FastLED.delay(1000 / FRAMES_PER_SECOND);


  //  EVERY_N_MILLISECONDS( 20 ) {    // was 20
  //    gHue++;  // slowly cycle the "base color" through the rainbow
  //  }
  EVERY_N_SECONDS( 169 ) {
    gHue++;  // slowly cycle base color
  }
}

void prepareLEDtime() {
  // Het is
  fill_solid (LEDs + 0, 3, gHue);
  fill_solid (LEDs + 4, 2, gHue);
  // minutes
  byte minutes = currentDateTime.minute;
  if (minutes >= 0 && minutes <= 4) {
    // heel uur
    fill_solid (LEDs + 47, 1, gHue);
    fill_solid (LEDs + 104, 2, gHue);
    fillminutedetail(0 + minutes);
  } else if (minutes >= 5 && minutes <= 9) {
    // vijf
    fillvijf();
    // over
    fillover();
    fillminutedetail(minutes - 5);
  } else if (minutes >= 10 && minutes <= 14) {
    // tien
    filltien();
    // over
    fillover();
    fillminutedetail(minutes - 10);
  } else if (minutes >= 15 && minutes <= 19) {
    // kwart
    fillkwart();
    // over
    fillover();
    fillminutedetail(minutes - 15);
  } else if (minutes >= 20 && minutes <= 24) {
    // tien
    filltien();
    // voor
    fillvoor();
    fillhalf();
    fillminutedetail(minutes - 20);
  } else if (minutes >= 25 && minutes <= 29) {
    fillvijf();
    fillvoor();
    fillhalf();
    fillminutedetail(minutes - 25);
  } else if (minutes >= 30 && minutes <= 34) {
    // half
    fillhalf();
    fillminutedetail(minutes - 30);
  } else if (minutes >= 35 && minutes <= 39) {
    fillvijf();
    fillover();
    fillhalf();
    fillminutedetail(minutes - 35);
  } else if (minutes >= 40 && minutes <= 44) {
    filltien();
    fillover();
    fillhalf();
    fillminutedetail(minutes - 40);
  } else if (minutes >= 45 && minutes <= 49) {
    fillkwart();
    fillvoor();
    fillminutedetail(minutes - 45);
  } else if (minutes >= 50 && minutes <= 54) {
    filltien();
    fillvoor();
    fillminutedetail(minutes - 50);
  } else if (minutes >= 55 && minutes <= 59) {
    fillvijf();
    fillvoor();
    fillminutedetail(minutes - 55);
  }
  // hour
  byte hours = currentDateTime.hour;
  // dagdeel
  if (hours >= 0 && hours <= 5) {
    // nacht
    fill_solid (LEDs + 60, 4, gHue);
    fill_solid (LEDs + 120, 1, gHue);
  } else if (hours >= 6 && hours <= 11) {
    // ochtend
    fill_solid (LEDs + 121, 7, gHue);
  } else if (hours >= 12 && hours <= 17) {
    // middag
    fill_solid (LEDs + 53, 3, gHue);
    fill_solid (LEDs + 112, 3, gHue);
  } else if (hours >= 18 && hours <= 23) {
    // avond
    fill_solid (LEDs + 115, 5, gHue);
  }
  // uren
  if (hours == 0) hours = 12;
  if (minutes >= 20 && minutes <= 59) hours++;
  if (hours > 12) hours = hours - 12;
  switch (hours) {
    case 1:  //een
      if (versie == 3) {
        fill_solid (LEDs + 81, 3, gHue);
      } else {
        fill_solid (LEDs + 23, 1, gHue);
        fill_solid (LEDs + 80, 2, gHue);
      }
      break;
    case 2:   // twee
      fill_solid (LEDs + 21, 3, gHue);
      fill_solid (LEDs + 80, 1, gHue);
      break;
    case 3:
      if (versie == 3) {
        fill_solid (LEDs + 84, 4, gHue);
      } else {
        fill_solid (LEDs + 83, 4, gHue);
      }
      break;
    case 4:
      fill_solid (LEDs + 24, 4, gHue);
      break;
    case 5:
      fill_solid (LEDs + 28, 4, gHue);
      break;
    case 6:
      fill_solid (LEDs + 88, 3, gHue);
      break;
    case 7:
      fill_solid (LEDs + 91, 5, gHue);
      break;
    case 8:
      fill_solid (LEDs + 32, 4, gHue);
      break;
    case 9:
      fill_solid (LEDs + 36, 4, gHue);
      fill_solid (LEDs + 96, 1, gHue);
      break;
    case 10:
      fill_solid (LEDs + 97, 4, gHue);
      break;
    case 11:
      fill_solid (LEDs + 101, 3, gHue);
      break;
    case 12:
      fill_solid (LEDs + 40, 6, gHue);
      break;
    default:
      // nothing
      break;
  }
  // in de
  fill_solid (LEDs + 107, 2, gHue);
  fill_solid (LEDs + 110, 2, gHue);
}

void fillvijf () {
  fill_solid (LEDs + 7, 1, gHue);
  fill_solid (LEDs + 64, 3, gHue);
}

void fillover() {
  fill_solid (LEDs + 76, 4, gHue);
}

void filltien () {
  fill_solid (LEDs + 68, 4, gHue);
}

void fillkwart () {
  fill_solid (LEDs + 8, 5, gHue);
}

void fillvoor () {
  fill_solid (LEDs + 15, 1, gHue);
  fill_solid (LEDs + 72, 3, gHue);
}

void fillhalf () {
  fill_solid (LEDs + 16, 4, gHue);
}

void fillminutedetail (byte minutedetail) {
  if (minutedetail > 0) {
    fill_solid (LEDs + 56, minutedetail, gHue);
  }
}

void speaktime() {
  byte minutedetail;
  // Het is
  speeltrack(310, true);
  // minutes
  byte minutes = currentDateTime.minute;
  if (minutes >= 0 && minutes <= 4) {
    // heel uur
    // skip this
    minutedetail = 0 + minutes;
  } else if (minutes >= 5 && minutes <= 9) {
    // vijf
    speeltrack(1005, true);
    // over
    speeltrack(304, true);
    minutedetail = minutes - 5;
  } else if (minutes >= 10 && minutes <= 14) {
    // tien
    speeltrack(1010, true);
    // over
    speeltrack(304, true);
    minutedetail = minutes - 10;
  } else if (minutes >= 15 && minutes <= 19) {
    // kwart
    speeltrack(302, true);
    // over
    speeltrack(304, true);
    minutedetail = minutes - 15;
  } else if (minutes >= 20 && minutes <= 24) {
    // tien
    speeltrack(1010, true);
    // voor
    speeltrack(303, true);
    speeltrack(305, true);
    minutedetail = minutes - 20;
  } else if (minutes >= 25 && minutes <= 29) {
    speeltrack(1005, true);
    speeltrack(303, true);
    speeltrack(305, true);
    minutedetail = minutes - 25;
  } else if (minutes >= 30 && minutes <= 34) {
    // half
    speeltrack(305, true);
    minutedetail = minutes - 30;
  } else if (minutes >= 35 && minutes <= 39) {
    speeltrack(1005, true);
    speeltrack(304, true);
    speeltrack(305, true);
    minutedetail = minutes - 35;
  } else if (minutes >= 40 && minutes <= 44) {
    speeltrack(1010, true);
    speeltrack(304, true);
    speeltrack(305, true);
    minutedetail = minutes - 40;
  } else if (minutes >= 45 && minutes <= 49) {
    speeltrack(302, true);
    speeltrack(303, true);
    minutedetail = minutes - 45;
  } else if (minutes >= 50 && minutes <= 54) {
    speeltrack(1010, true);
    speeltrack(303, true);
    minutedetail = minutes - 50;
  } else if (minutes >= 55 && minutes <= 59) {
    speeltrack(1005, true);
    speeltrack(303, true);
    minutedetail = minutes - 55;
  }
  // uren
  byte hours = currentDateTime.hour;
  if (hours == 0) hours = 12;
  if (minutes >= 20 && minutes <= 59) hours++;
  if (hours > 12) hours = hours - 12;
  speeltrack(1000 + hours, true);
  if (minutes >= 0 && minutes <= 4) {
    // heel uur
    speeltrack(301, true);
  }
  // in de
  speeltrack(340, true);
  hours = currentDateTime.hour;
  // dagdeel
  if (hours >= 0 && hours <= 5) {
    // nacht
    speeltrack(343, true);
  } else if (hours >= 6 && hours <= 11) {
    // ochtend
    speeltrack(344, true);
  } else if (hours >= 12 && hours <= 17) {
    // middag
    speeltrack(341, true);
  } else if (hours >= 18 && hours <= 23) {
    // avond
    speeltrack(342, true);
  }
  // en minute detail
  if (minutedetail > 0) {
    // en
    speeltrack(350, true);
    speeltrack(1000 + minutedetail, true);
    if (minutedetail == 1) {
      speeltrack(352, true);   // minute
    } else {
      speeltrack(353, true);   // minuten
    }
  }
}



void startWiFi() {
  wifiMulti.addAP(ssid, pass);
  Serial.println("Connecting");
  byte i = 0;
  LEDs[i++] = CRGB::Green;
  FastLED.show();
  while (wifiMulti.run() != WL_CONNECTED) {
    delay(250);
    Serial.print('.');
    LEDs[i++] = CRGB::Green;
    FastLED.show();
  }
  Serial.println("\r\n");
  Serial.print("Connected to ");
  Serial.println(WiFi.SSID());
  Serial.print("IP address:\t");
  Serial.print(WiFi.localIP());
  Serial.println("\r\n");
}

void startUDP() {
  Serial.println("Starting UDP");
  UDP.begin(123);                          // Start listening for UDP messages on port 123
  Serial.print("Local port:\t");
  Serial.println(UDP.localPort());
  Serial.println();
}

uint32_t getTime() {
  if (UDP.parsePacket() == 0) { // If there's no response (yet)
    return 0;
  }
  UDP.read(NTPBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
  // Combine the 4 timestamp bytes into one 32-bit number
  uint32_t NTPTime = (NTPBuffer[40] << 24) | (NTPBuffer[41] << 16) | (NTPBuffer[42] << 8) | NTPBuffer[43];
  // Convert NTP time to a UNIX timestamp:
  // Unix time starts on Jan 1 1970. That's 2208988800 seconds in NTP time:
  const uint32_t seventyYears = 2208988800UL;
  // subtract seventy years:
  uint32_t UNIXTime = NTPTime - seventyYears;
  return UNIXTime;
}

void sendNTPpacket(IPAddress& address) {
  memset(NTPBuffer, 0, NTP_PACKET_SIZE);  // set all bytes in the buffer to 0
  // Initialize values needed to form NTP request
  NTPBuffer[0] = 0b11100011;   // LI, Version, Mode
  // send a packet requesting a timestamp:
  UDP.beginPacket(address, 123); // NTP requests are to port 123
  UDP.write(NTPBuffer, NTP_PACKET_SIZE);
  UDP.endPacket();
}

void convertTime(uint32_t time) {
  // Correct time zone
  time += (3600 * timeZone);
  currentDateTime.second = time % 60;
  currentDateTime.minute = time / 60 % 60;
  currentDateTime.hour   = time / 3600 % 24;
  time  /= 60;  // To minutes
  time  /= 60;  // To hours
  time  /= 24;  // To days
  currentDateTime.dayofweek = ((time + 4) % 7) + 1;
  int year = 0;
  int days = 0;
  while ((unsigned)(days += (LEAP_YEAR(year) ? 366 : 365)) <= time) {
    year++;
  }
  days -= LEAP_YEAR(year) ? 366 : 365;
  time  -= days; // To days in this year, starting at 0
  days = 0;
  byte month = 0;
  byte monthLength = 0;
  for (month = 0; month < 12; month++) {
    if (month == 1) { // February
      if (LEAP_YEAR(year)) {
        monthLength = 29;
      } else {
        monthLength = 28;
      }
    } else {
      monthLength = monthDays[month];
    }
    if (time >= monthLength) {
      time -= monthLength;
    } else {
      break;
    }
  }
  currentDateTime.day = time + 1;
  currentDateTime.year = year + 1970;
  currentDateTime.month = month + 1;
  // Correct European Summer time
  if (summerTime()) {
    currentDateTime.hour += 1;
  }

#ifdef DEBUG_ON
  Serial.print(currentDateTime.year);
  Serial.print(" ");
  Serial.print(currentDateTime.month);
  Serial.print(" ");
  Serial.print(currentDateTime.day);
  Serial.print(" ");
  Serial.print(currentDateTime.hour);
  Serial.print(" ");
  Serial.print(currentDateTime.minute);
  Serial.print(" ");
  Serial.print(currentDateTime.second);
  Serial.print(" day of week: ");
  Serial.print(currentDateTime.dayofweek);
  Serial.print(" summer time: ");
  Serial.print(summerTime());
  Serial.println();
#endif
}

boolean summerTime() {
  if (currentDateTime.month < 3 || currentDateTime.month > 10) return false;  // No summer time in Jan, Feb, Nov, Dec
  if (currentDateTime.month > 3 && currentDateTime.month < 10) return true;   // Summer time in Apr, May, Jun, Jul, Aug, Sep
  if (currentDateTime.month == 3 && (currentDateTime.hour + 24 * currentDateTime.day) >= (3 +  24 * (31 - (5 * currentDateTime.year / 4 + 4) % 7)) || currentDateTime.month == 10 && (currentDateTime.hour + 24 * currentDateTime.day) < (3 +  24 * (31 - (5 * currentDateTime.year / 4 + 1) % 7)))
    return true;
  else
    return false;
}

void applyMap() {
  maplength = 0;
  for (int i = 0 ; i < NUM_LEDS; i++) {
    if (LEDs[Map[i]]) {   // test if not completely black
      // map index
      mapindex[maplength] = Map[i];
      maplength++;
    }
  }
}

void applycoloreffect(byte effect) {
  // 1 = random;
  // 2 = addGlitter(80);
  // 3 = bpm();
  // 4 = sinelon();
  // 5 = juggle();
  // 6 = confetti();
  switch (effect) {
    case 2:
      addGlitter(80);
      break;
    case 3:
      bpm();
      break;
    case 4:
      sinelon();
      break;
    case 5:
      juggle();
      break;
    case 6:
      confetti();
      break;
    default:
      break;
  }

}


void rainbow() {
  byte startgHue = gHue;
  //  Serial.print("apply rainbow:");
  for (int i = 0 ; i < maplength; i++) {
    LEDs[mapindex[i]].setHue(startgHue);
    startgHue += deltaHue;
  }
}

void rainbowWithGlitter()
{
  // built-in FastLED rainbow, plus some random sparkly glitter
  rainbow();
  addGlitter(80);
}

void addGlitter( fract8 chanceOfGlitter)
{
  if ( random8() < chanceOfGlitter) {
    LEDs[mapindex[ random16(maplength) ]] += CRGB::White;
  }
}

void confetti()
{
  // random colored speckles that blink in and fade smoothly
  fadeToBlackBy( LEDs, NUM_LEDS, 10);
  int pos = random16(maplength);
  LEDs[mapindex[pos]] += CHSV( gHue + random8(64), 200, 255);
}

void sinelon()
{
  // a colored dot sweeping back and forth, with fading trails
  fadeToBlackBy( LEDs, NUM_LEDS, 20);
  int pos = beatsin16( 13, 0, maplength - 1 );
  LEDs[mapindex[pos]] += CHSV( gHue, 255, 192);
}

void bpm()
{
  // colored stripes pulsing at a defined Beats-Per-Minute (BPM)
  uint8_t BeatsPerMinute = 62;
  CRGBPalette16 palette = PartyColors_p;
  uint8_t beat = beatsin8( BeatsPerMinute, 64, 255);
  for ( int i = 0; i < maplength; i++) { //9948
    LEDs[mapindex[i]] = ColorFromPalette(palette, gHue + (i * 2), beat - gHue + (i * 10));
  }
}

void juggle() {
  // eight colored dots, weaving in and out of sync with each other
  fadeToBlackBy( LEDs, NUM_LEDS, 20);
  byte dothue = 0;
  for ( int i = 0; i < 8; i++) {
    LEDs[mapindex[beatsin16( i + 7, 0, maplength - 1 )]] |= CHSV(dothue, 200, 255);
    dothue += 32;
  }
}

void speeltrack(int tracknbr, boolean wachten) {
  //  wdt_disable();    // prefend watchdog timer to reset ESP
  int busy = digitalRead(DFPlayerBusyPin);   // HIGH (1) = not busy  LOW (0) = busy
  while (busy == LOW && wachten) {
    //  while (DFPlayer.isPlaying() && wachten) {
    delay(150);
//       Serial.print(".");   //
    busy = digitalRead(DFPlayerBusyPin);
  }
//   Serial.println("s");  //
  DFPlayer.playFromMP3Folder(tracknbr);
  delay(500);    // give dfpalyer time to become busy, otherwise it will skip tracks
//  Serial.println("p");
  //  wdt_enable(WDTO_1S);     // Default WDT is 1 s
}
